from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics import faithfulness
from typing import List

"""
RAGAS Faithfulness Docs: https://docs.ragas.io/en/latest/concepts/metrics/faithfulness.html
RAGAS Faithfulness Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/_faithfulness.py
"""
class RagasFaithfulness(RagasEvaluator):
    """
    This measures the factual consistency of the generated response against the given context.
    """
    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_FAITHFULNESS.value

    @property
    def display_name(self):
        return "Ragas Faithfulness"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_FAITHFULNESS.value]
    
    @property
    def ragas_metric(self):
        return faithfulness
    
    @property
    def ragas_metric_name(self):
        return "faithfulness"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "contexts", "response"]

    @property
    def examples(self):
        return None
    
    @property
    def grade_reason(self) -> str:
        return "The generated answer is regarded as faithful if all the claims that are made in the answer can be inferred from the given context. To calculate this a set of claims from the generated answer is first identified. Then each one of these claims are cross checked with given context to determine if it can be inferred from given context or not"
    
    def generate_data_to_evaluate(self, contexts, query, response, **kwargs) -> dict:
        """
        Generates data for evaluation.

        :param context: list of strings of retrieved context
        :param query: user query
        :param response: llm response
        :return: A dictionary with formatted data for evaluation.
        """
        data = {
            "contexts": [contexts],
            "question": [query],
            "answer": [response]
        }
        return data



from athina.interfaces.model import Model
from ..ragas_evaluator import RagasEvaluator
from athina.evals.eval_type import RagasEvalTypeId
from athina.metrics.metric_type import MetricType
from ragas.metrics.critique import harmfulness
from typing import List

"""
RAGAS Harmfulness Docs: https://docs.ragas.io/en/latest/concepts/metrics/critique.html
RAGAS Harmfulness Github: https://github.com/explodinggradients/ragas/blob/main/src/ragas/metrics/critique.py
"""
class RagasHarmfulness(RagasEvaluator):
    """
    This measures if the generated response has the potential to cause harm to individuals, groups, or society at large
    """
    @property
    def name(self):
        return RagasEvalTypeId.RAGAS_HARMFULNESS.value

    @property
    def display_name(self):
        return "Ragas Harmfulness"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.RAGAS_HARMFULNESS.value]
    
    @property
    def ragas_metric(self):
        return harmfulness
    
    @property
    def ragas_metric_name(self):
        return "harmfulness"

    @property
    def default_model(self):
        return Model.GPT35_TURBO.value

    @property
    def required_args(self):
        return ["query", "contexts", "response", "expected_response"]

    @property
    def examples(self):
        return None
    
    @property
    def grade_reason(self) -> str:
        return "This is calculated by how much potential generated response has to cause harm to individuals, groups, or society at large"

    def generate_data_to_evaluate(self, contexts, query, response, expected_response, **kwargs) -> dict:
        """
        Generates data for evaluation.

        :param context: list of strings of retrieved context
        :param query: user query
        :param response: llm response
        :param expected_response: expected output
        :return: A dictionary with formatted data for evaluation.
        """
        data = {
            "contexts": [contexts],
            "question": [query],
            "answer": [response],
            "ground_truths": [[expected_response]]
        }
        return data



from athina.evals import Regex, ContainsAny, ContainsAll, Contains, ContainsNone, ContainsJson, ContainsEmail, IsJson, IsEmail, NoInvalidLinks, ContainsLink, ContainsValidLink, Equals, StartsWith, EndsWith, LengthLessThan, LengthGreaterThan, ApiCall, DoesResponseAnswerQuery, Faithfulness, BaseEvaluator, ContextContainsEnoughInformation, SummaryAccuracy, Groundedness, GradingCriteria, CustomPrompt, RagasContextRelevancy, RagasAnswerRelevancy, RagasAnswerCorrectness, RagasAnswerSemanticSimilarity, RagasCoherence, RagasConciseness, RagasContextPrecision, RagasContextRecall, RagasFaithfulness, RagasHarmfulness, RagasMaliciousness

function_operations = {
    "Regex": Regex,
    "ContainsAny": ContainsAny,
    "ContainsAll": ContainsAll,
    "Contains": Contains,
    "ContainsNone": ContainsNone,
    "ContainsJson": ContainsJson,
    "ContainsEmail": ContainsEmail,
    "IsJson": IsJson,
    "IsEmail": IsEmail,
    "NoInvalidLinks": NoInvalidLinks,
    "ContainsLink": ContainsLink,
    "ContainsValidLink": ContainsValidLink,
    "Equals": Equals,
    "StartsWith": StartsWith,
    "EndsWith": EndsWith,
    "LengthLessThan": LengthLessThan,
    "LengthGreaterThan": LengthGreaterThan,
    "ApiCall": ApiCall,
}

llm_operations = {
    "Draq": DoesResponseAnswerQuery,
    "Irftc": Faithfulness,
    "BaseEvaluator": BaseEvaluator,
    "Ccei": ContextContainsEnoughInformation,
    "SummaryAccuracy": SummaryAccuracy,
    "Groundedness": Groundedness,
    "GradingCriteria": GradingCriteria,
    "CustomPrompt": CustomPrompt,
}

ragas_operations = {
    "RagasContextRelevancy": RagasContextRelevancy,
    "RagasAnswerRelevancy": RagasAnswerRelevancy,
    "RagasAnswerCorrectness": RagasAnswerCorrectness,
    "RagasAnswerSemanticSimilarity": RagasAnswerSemanticSimilarity,
    "RagasCoherence": RagasCoherence,
    "RagasConciseness": RagasConciseness,
    "RagasContextPrecision": RagasContextPrecision,
    "RagasContextRecall": RagasContextRecall,
    "RagasFaithfulness": RagasFaithfulness,
    "RagasHarmfulness": RagasHarmfulness,
    "RagasMaliciousness": RagasMaliciousness
}


def get_evaluator(evaluator_type):
    if evaluator_type in function_operations:
        return function_operations[evaluator_type]
    elif evaluator_type in llm_operations:
        return llm_operations[evaluator_type]
    elif evaluator_type in ragas_operations:
        return ragas_operations[evaluator_type]
    else:
        raise ValueError(f"Invalid evaluator type: {evaluator_type}")
    




from typing import List, Optional

from athina.llms.abstract_llm_service import AbstractLlmService
from ..llm_evaluator import LlmEvaluator
from athina.metrics.metric_type import MetricType
from athina.evals.eval_type import LlmEvalTypeId


class GradingCriteria(LlmEvaluator):
    """
    This evaluator checks if the response is correct according to a provided `grading_criteria`.
    """

    _examples = []

    def __init__(self, 
        grading_criteria: str, 
        model: Optional[str] = None,
        llm_service: Optional[AbstractLlmService] = None
    ):
        if grading_criteria is None:
            raise Exception("Eval is incorrectly configured: grading_criteria is required for GradingCriteria evaluator")
        super().__init__(
            model=model,
            grading_criteria=grading_criteria,
            llm_service=llm_service
        )

    @property
    def name(self):
        return LlmEvalTypeId.GRADING_CRITERIA.value

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.PASSED.value]

    @property
    def display_name(self):
        return "Response matches Grading Criteria"

    @property
    def default_model(self):
        return "gpt-4-1106-preview"

    @property
    def required_args(self):
        return ["response"]

    @property
    def examples(self):
        return self._examples



from typing import List, Optional
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint as BaseDataPoint
from .base_loader import Loader as BaseLoader


class DataPoint(BaseDataPoint):
    """Data point for a single inference."""

    query: Optional[str]
    context: Optional[str]
    response: Optional[str]
    expected_response: Optional[str]


class Loader(BaseLoader):
    """
    This class is a generic data loader for evals

    Attributes:
        col_query (str): The column name corresponding to the user's query.
        col_context (str): The column name corresponding to the retrieved context.
        col_response (str): The column name corresponding to the response.
        col_expected_response (str): The column name corresponding to the expected response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with queries, context, response and other attributes if present.
    """

    def __init__(
        self,
        col_query="query",
        col_context="context",
        col_response="response",
        col_expected_response="expected_response",
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_query = col_query
        self.col_context = col_context
        self.col_response = col_response
        self.col_expected_response = col_expected_response
        self._raw_dataset = {}
        self._processed_dataset: List[DataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.
        """
        for raw_instance in self._raw_dataset:
            # Create a processed instance with mandatory fields
            processed_instance = {}
            # add only if the key is present in the raw instance
            if self.col_query in raw_instance:
                processed_instance["query"] = raw_instance[self.col_query]
            if self.col_context in raw_instance:
                processed_instance["context"] = raw_instance[self.col_context]
            if self.col_response in raw_instance:
                processed_instance["response"] = raw_instance[self.col_response]
            if self.col_expected_response in raw_instance:
                processed_instance["expected_response"] = raw_instance[self.col_expected_response]
            # Store the results
            processed_data_point = DataPoint(**processed_instance)
            self._processed_dataset.append(processed_data_point)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: int = 10,
        context_key: Optional[str] = None,
    ):
        """
        Load data from Athina API.
        By default, this will fetch the last 10 inferences from the API.
        """
        pass


from enum import Enum


class ConversationEvalTypeId(Enum):
    CONVERSATION_RESOLUTION = "ConversationResolution"
    CONVERSATION_COHERENCE = "ConversationCoherence"


class LlmEvalTypeId(Enum):
    CONTEXT_CONTAINS_ENOUGH_INFORMATION = "Ccei"
    DOES_RESPONSE_ANSWER_QUERY = "Draq"
    FAITHFULNESS = "Irftc"
    GRADING_CRITERIA = "GradingCriteria"
    CUSTOM_PROMPT = "CustomPrompt"
    SUMMARY_ACCURACY = "SummaryAccuracy"


class RagasEvalTypeId(Enum):
    RAGAS_CONTEXT_RELEVANCY = "RagasContextRelevancy"
    RAGAS_ANSWER_RELEVANCY = "RagasAnswerRelevancy"
    RAGAS_CONTEXT_PRECISION = "RagasContextPrecision"
    RAGAS_FAITHFULNESS = "RagasFaithfulness"
    RAGAS_CONTEXT_RECALL = "RagasContextRecall"
    RAGAS_ANSWER_SEMANTIC_SIMILARITY = "RagasAnswerSemanticSimilarity"
    RAGAS_ANSWER_CORRECTNESS = "RagasAnswerCorrectness"
    RAGAS_HARMFULNESS = "RagasHarmfulness"
    RAGAS_MALICIOUSNESS = "RagasMaliciousness"
    RAGAS_COHERENCE = "RagasCoherence"
    RAGAS_CONCISENESS = "RagasConciseness"


class FunctionEvalTypeId(Enum):
    REGEX = "Regex"
    CONTAINS_ANY = "ContainsAny"
    CONTAINS_ALL = "ContainsAll"
    CONTAINS = "Contains"
    CONTAINS_NONE = "ContainsNone"
    CONTAINS_JSON = "ContainsJson"
    CONTAINS_EMAIL = "ContainsEmail"
    IS_JSON = "IsJson"
    IS_EMAIL = "IsEmail"
    NO_INVALID_LINKS = "NoInvalidLinks"
    CONTAINS_LINK = "ContainsLink"
    CONTAINS_VALID_LINK = "ContainsValidLink"
    EQUALS = "Equals"
    STARTS_WITH = "StartsWith"
    ENDS_WITH = "EndsWith"
    LENGTH_LESS_THAN = "LengthLessThan"
    LENGTH_GREATER_THAN = "LengthGreaterThan"
    API_CALL = "ApiCall"


def is_llm_eval(evaluator_type: str) -> bool:
    return any(evaluator_type == member.value for member in LlmEvalTypeId)


def is_ragas_eval(evaluator_type: str) -> bool:
    return any(evaluator_type == member.value for member in RagasEvalTypeId)


def is_function_eval(evaluator_type: str) -> bool:
    return any(evaluator_type == member.value for member in FunctionEvalTypeId)



from typing import List, TypedDict, Optional
from athina.helpers.athina_logging_helper import AthinaLoggingHelper
from athina.evals.llm.llm_evaluator import LlmEvaluator
from athina.interfaces.result import EvalResult, BatchRunResult
from athina.interfaces.data import DataPoint
from athina.interfaces.athina import AthinaExperiment
from athina.services.athina_api_service import AthinaApiService
import pandas as pd
import json
import hashlib


class DataPointWithEvalResults(TypedDict):
    """A data point with its evaluation results."""

    data_point: DataPoint
    eval_results: List[EvalResult]


class LlmEvaluatorDescription(TypedDict):
    """A description of an LLM evaluator."""

    name: str
    display_name: str


class LlmBatchEvalResult(TypedDict):
    """Result of running a batch of LLM evaluations."""

    results: List[EvalResult]
    total_runtime: float
    passed_evals: int
    failed_evals: int
    total_evals: int
    total_datapoints: int


class EvalRunner:
    @staticmethod
    def eval_results_link(eval_request_id: str):
        return f"https://app.athina.ai/develop/request/{eval_request_id}"

    @staticmethod
    def flatten_eval_results(batch_eval_results):
        # Flatten the list of lists into a single list of evaluation results
        flattened_results = [item for sublist in batch_eval_results for item in sublist]
        return flattened_results

    @staticmethod
    def to_df(batch_eval_results):
        # Initialize a dictionary to hold the aggregated data
        aggregated_data = {}
        flattened_results = EvalRunner.flatten_eval_results(
            batch_eval_results=batch_eval_results
        )

        # Process each evaluation result
        for eval_result in flattened_results:
            # Serialize and hash the datapoint dictionary to create a unique identifier
            datapoint_hash = hashlib.md5(
                json.dumps(eval_result["data"], sort_keys=True).encode()
            ).hexdigest()

            # Initialize the datapoint in the aggregated data if not already present
            if datapoint_hash not in aggregated_data:
                aggregated_data[datapoint_hash] = eval_result[
                    "data"
                ]  # Include datapoint details

            # Update the aggregated data with metrics from this evaluation
            for metric in eval_result["metrics"]:
                metric_name = metric["id"]
                metric_value = metric["value"]
                aggregated_data[datapoint_hash][eval_result["display_name"] + " " + metric_name] = metric_value

        # Convert the aggregated data into a DataFrame
        df = pd.DataFrame(list(aggregated_data.values()))

        return df

    @staticmethod
    def batch_eval_result(
        eval_results: List[EvalResult],
    ) -> LlmBatchEvalResult:
        """
        Calculate metrics for a batch of LLM evaluations.

        Args:
            datapoints_with_eval_results: A list of DataPointWithEvalResults objects.

        Returns:
            A LlmBatchEvalResult object.
        """
        total_runtime = 0
        passed_evals = 0
        failed_evals = 0
        total_evals = 0

        # Iterate through each DataPointWithEvalResults
        for eval_result in eval_results:
            total_evals += 1
            total_runtime += eval_result.get("runtime", 0)

            # Counting passed and failed evaluations
            if eval_result.get("failure"):
                failed_evals += 1
            else:
                passed_evals += 1

        total_datapoints = len(eval_result)

        return LlmBatchEvalResult(
            results=eval_result,
            total_runtime=total_runtime,
            passed_evals=passed_evals,
            failed_evals=failed_evals,
            total_evals=total_evals,
            total_datapoints=total_datapoints,
        )

    @staticmethod
    def run_suite(
        evals: List[LlmEvaluator],
        data: List[DataPoint],
        experiment: Optional[AthinaExperiment] = None,
        max_parallel_evals: int = 1,
    ) -> List[LlmBatchEvalResult]:
        """
        Run a suite of LLM evaluations against a dataset.

        Args:
            evals: A list of LlmEvaluator objects.
            data: A list of data points.

        Returns:
            A list of LlmBatchEvalResult objects.
        """
        # Create eval request
        eval_suite_name = "llm_eval_suite" + "_" + ",".join(eval.name for eval in evals)
        eval_request_id = AthinaLoggingHelper.create_eval_request(
            eval_name=eval_suite_name,
            request_data={"data": data},
            request_type="suite",
        )

        # Log experiment
        if experiment is not None:
            AthinaLoggingHelper.log_experiment(
                eval_request_id=eval_request_id,
                experiment=experiment,
            )

        AthinaApiService.log_usage(eval_name=eval_suite_name, run_type="suite")

        batch_results = []
        for eval in evals:
            # Validate the dataset against the required args
            eval._validate_batch_args(data)

            # Run the evaluations
            if max_parallel_evals > 1:
                eval_results = eval._run_batch_generator_async(data, max_parallel_evals)
            else:
                eval_results = list(eval._run_batch_generator(data))

            # Log evaluation results to Athina
            AthinaLoggingHelper.log_eval_results(
                eval_request_id=eval_request_id,
                eval_results=eval_results,
            )

            batch_results.append(eval_results)

        print(
            f"You can view the evaluation results at {EvalRunner.eval_results_link(eval_request_id)}"
        )
        return EvalRunner.to_df(batch_results)



# Guardrails Evaluator
# https://hub.guardrailsai.com/validator/guardrails/gibberish_text

import math
from abc import abstractmethod
from typing import Optional, List
from athina.interfaces.athina import AthinaExperiment
from athina.interfaces.model import Model
import time
from typing import Optional, Any
from athina.interfaces.result import EvalResult, EvalResultMetric
from athina.interfaces.model import Model
from athina.helpers.logger import logger
from ...base_evaluator import BaseEvaluator
from datasets import Dataset
from athina.keys import OpenAiApiKey
from guardrails.hub import GibberishText as GuardrailsGibberishText
from guardrails import Guard
from athina.metrics.metric_type import MetricType


# Passes when the text is sensible, fails when the text is gibberish.
class GibberishText(BaseEvaluator):
    _validation_method: str
    _threshold: float

    def __init__(
        self,
        validation_method: str = "sentence",
        threshold: float = 0.75,
    ):
        self._validation_method = validation_method
        self._threshold = threshold

    @property
    def name(self) -> str:
        return "GibberishText"

    @property
    def display_name(self) -> str:
        return "Gibberish Text"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.SENSIBLE_TEXT.value]

    @property
    def required_args(self) -> List[str]:
        return ["response"]  # TODO: allow running this on user_query OR response

    @property
    def examples(self):
        pass

    def _evaluate(self, **kwargs) -> EvalResult:
        """
        Run the Guardrails evaluator.
        """
        start_time = time.time()
        self.validate_args(**kwargs)
        metrics = []
        try:

            text = kwargs["response"]

            # Initialize Validator
            val = GuardrailsGibberishText(
                threshold=self._threshold,
                validation_method=self._validation_method,
                on_fail="noop",
            )

            # Setup Guard
            guard = Guard.from_string(validators=[val])

            # Pass LLM output through guard
            guard_result = guard.parse(text)
            passed = guard_result.validation_passed
            grade_reason = "Text is sensible" if passed else "Text is gibberish"

            # Boolean evaluator
            metrics.append(
                EvalResultMetric(
                    id=MetricType.SENSIBLE_TEXT.value,
                    value=guard_result.validation_passed,
                )
            )

        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            raise e

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        llm_eval_result = EvalResult(
            name=self.name,
            display_name=self.display_name,
            data=kwargs,
            failure=None,
            reason=grade_reason,
            runtime=eval_runtime_ms,
            model=None,
            metrics=metrics,
        )
        return {k: v for k, v in llm_eval_result.items() if v is not None}



from typing import List
from ..llm_evaluator import LlmEvaluator
from .examples import CONTEXT_CONTAINS_ENOUGH_INFORMATION_EXAMPLES
from athina.evals.eval_type import LlmEvalTypeId
from athina.metrics.metric_type import MetricType


class ContextContainsEnoughInformation(LlmEvaluator):
    """
    This evaluator checks if the user's query can be answered using only the information in the context.
    """

    SYSTEM_MESSAGE_TEMPLATE = """
    You are an expert at evaluating whether a chatbot can answer a user's query using ONLY the information provided to you as context.
    You are not concerned with factual correctness or accuracy. You only care whether the context contains enough information to answer the user's query.
    """

    USER_MESSAGE_TEMPLATE = """
        Let's think step by step:

        1. Consider the following: 
        user's query: {query}.
        context: {context}.
        2. Determine if the chatbot can answer the user's query with nothing but the "context" information provided to you.
        3. Provide a brief explanation of why the context does or does not contain sufficient information, labeled as 'explanation', leading up to a verdict (Pass/Fail) labeled as 'result'.
        4. Always return a JSON object in the following format: "result": 'result', "explanation": 'explanation'.

        Here's are some examples: 
        {examples}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def name(self):
        return LlmEvalTypeId.CONTEXT_CONTAINS_ENOUGH_INFORMATION.value

    @property
    def display_name(self):
        return "Context Contains Enough Information"

    @property
    def metric_ids(self) -> List[str]:
        return [MetricType.PASSED.value]

    @property
    def default_model(self):
        return "gpt-4-1106-preview"

    @property
    def required_args(self):
        return ["query", "context"]

    @property
    def examples(self):
        return CONTEXT_CONTAINS_ENOUGH_INFORMATION_EXAMPLES

    def _user_message(
        self,
        query: str,
        context: str,
        **kwargs,
    ) -> str:
        return self.USER_MESSAGE_TEMPLATE.format(
            query=query,
            context=context,
            examples=self.examples,
        )



from typing import List, Optional
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint
from .base_loader import Loader


class RagDataPoint(DataPoint):
    """Data point for a single RAG inference."""

    query: str
    context: str
    response: str


class RagLoader(Loader):
    """
    This class is a data loader for retrieval augmented generation (RAG) datasets.

    Attributes:
        col_query (str): The column name corresponding to the user's query.
        col_context (str): The column name corresponding to the retrieved context.
        col_response (str): The column name corresponding to the response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with queries, context, response and other attributes if present.
    """

    def __init__(
        self,
        col_query="query",
        col_context="context",
        col_response="response",
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_query = col_query
        self.col_context = col_context
        self.col_response = col_response
        self._raw_dataset = {}
        self._processed_dataset: List[RagDataPoint] = []

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.

        Raises:
            KeyError: If mandatory columns (query, context or response) are missing in the raw dataset.
        """
        for raw_instance in self._raw_dataset:
            # Check for mandatory columns in raw_instance
            if self.col_query not in raw_instance:
                raise KeyError(f"'{self.col_query}' not found in provided data.")
            if self.col_context not in raw_instance:
                raise KeyError(f"'{self.col_context}' not found in provided data.")
            if self.col_response not in raw_instance:
                raise KeyError(f"'{self.col_response}' not found in provided data.")
            

            # Create a processed instance with mandatory fields
            processed_instance = {
                "query": raw_instance[self.col_query],
                "context": raw_instance[self.col_context],
                "response": raw_instance[self.col_response],
            }


            # Store the results
            self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: int = 10,
        context_key: Optional[str] = None,
    ):
        """
        Load data from Athina API.
        By default, this will fetch the last 10 inferences from the API.
        """
        pass


from typing import Union
from .metric import Metric


class Passed(Metric):
    """
    Boolean metric indicating whether the evaluation passed the specified criteria.
    """

    @staticmethod
    def compute(passed: Union[int, bool]):
        """
        Computes the result.

        Returns:
            bool: Whether the evaluation passed or not.
        """
        return bool(passed)



from openai import OpenAI
from retrying import retry
from timeout_decorator import timeout
from athina.helpers.json import JsonHelper
from athina.keys import OpenAiApiKey
from athina.interfaces.model import Model
from athina.errors.exceptions import NoOpenAiApiKeyException
from .abstract_llm_service import AbstractLlmService

DEFAULT_TEMPERATURE = 0.0


class OpenAiService(AbstractLlmService):
    _instance = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(OpenAiService, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        openai_api_key = OpenAiApiKey.get_key()
        if openai_api_key is None:
            raise NoOpenAiApiKeyException()
        self.openai = OpenAI(api_key=openai_api_key)

    def embeddings(self, text: str) -> list:
        """
        Fetches response from OpenAI's Embeddings API.
        """
        try:
            response = self.openai.embeddings.create(
                model="text-embedding-ada-002",
                input=text,
                encoding_format="float"
            )
            return response.data[0].embedding
        except Exception as e:
            print(f"Error in Embeddings: {e}")
            raise e

    @retry(stop_max_attempt_number=3, wait_fixed=2000)
    def chat_completion(self, messages, model, temperature=DEFAULT_TEMPERATURE) -> str:
        """
        Fetches response from OpenAI's ChatCompletion API.
        """
        try:
            response = self.openai.chat.completions.create(
                model=model, messages=messages, temperature=temperature
            )
            return response.choices[0].message.content
        except Exception as e:
            print(f"Error in ChatCompletion: {e}")
            raise e

    @retry(stop_max_attempt_number=3, wait_fixed=2000)
    def chat_completion_json(self, messages, model, temperature=DEFAULT_TEMPERATURE) -> str:
        """
        Fetches response from OpenAI's ChatCompletion API using JSON mode.
        """
        try:
            response = self.openai.chat.completions.create(
                model=model,
                messages=messages,
                temperature=temperature,
                response_format={"type": "json_object"},
            )
            return response.choices[0].message.content
        except Exception as e:
            print(f"Error in ChatCompletion: {e}")
            raise e

    def json_completion(self, messages, model, temperature=DEFAULT_TEMPERATURE):
        """
        Fetches response from OpenAI's ChatCompletion API using JSON mode.
        """
        try:
            if Model.supports_json_mode(model):
                chat_completion_response = self.chat_completion_json(
                    model=model,
                    messages=messages,
                    temperature=temperature,
                )
            else:
                chat_completion_response = self.chat_completion(
                    model=model,
                    messages=messages,
                    temperature=temperature,
                )

            # Extract JSON object from LLM response
            return JsonHelper.extract_json_from_text(chat_completion_response)
        
        except Exception as e:
            print(f"Error in ChatCompletion: {e}")
            raise e


from typing import List, Optional, TypedDict
from athina.interfaces.athina import AthinaFilters
from athina.interfaces.data import DataPoint
from .base_loader import Loader
from llama_index.indices.query.base import BaseQueryEngine


class RagasDataPoint(TypedDict):
    """Data point for a single RAG inference."""

    query: Optional[str]
    contexts: Optional[List[str]]
    response: Optional[str]
    expected_response: Optional[str]


class RagasLoader(Loader):
    """
    This class is a data loader for retrieval augmented generation (RAG) datasets.

    Attributes:
        col_query (str): The column name corresponding to the user's query.
        col_contexts (List[str]): The column name corresponding to the retrieved contexts.
        col_response (str): The column name corresponding to the response.
        col_expected_response (str): The column name corresponding to the expected response.
        raw_dataset (dict): The raw dataset as loaded from the source.
        processed_dataset (list): The processed dataset with queries, context, response and other attributes if present.
    """

    def __init__(
        self,
        col_query="query",
        col_contexts="contexts",
        col_response="response",
        col_expected_response="expected_response",
        query_engine: Optional[BaseQueryEngine] = None,
    ):
        """
        Initializes the loader with specified or default column names.
        """
        self.col_query = col_query
        self.col_contexts = col_contexts
        self.col_response = col_response
        self.col_expected_response = col_expected_response
        self.query_engine = query_engine
        self._raw_dataset = {}
        self._processed_dataset: List[RagasDataPoint] = []

    def _fetch_context_and_response_for_llama_index(self, query: str):
        """
        Fetches the context and response from the llama index query engine.
        """
        try:
            from llama_index.async_utils import run_async_tasks
        except ImportError:
            raise ImportError(
            "llama_index must be installed to use this function. "
            "Install it with `pip install llama_index`."
            )
        contexts = []
        query_engine_response = self.query_engine.query(query)
        response = query_engine_response.response
        for c in query_engine_response.source_nodes:
            text = c.node.get_text()
            contexts.append(text)

        return contexts, response

    def _generate_processed_instance_for_llama_index(self, raw_instance: dict) -> RagasDataPoint:
        """
        Generates a processed instance for the llama index query engine.
        """
        if self.col_query not in raw_instance:
            raise ValueError(f"'{self.col_query}' not found in provided data.")
        if self.col_query in raw_instance and not isinstance(raw_instance.get(self.col_query), str):
            raise TypeError(f"'{self.col_query}' is not of type string.")
        if self.col_expected_response in raw_instance and not isinstance(raw_instance.get(self.col_expected_response), str):
            raise TypeError(f"'{self.col_expected_response}' is not of type string.")

        contexts, response = self._fetch_context_and_response_for_llama_index(raw_instance.get(self.col_query))
        processed_instance = {
            "query": raw_instance.get(self.col_query),
            "contexts": contexts,
            "response": response,
            "expected_response": raw_instance.get(self.col_expected_response, None)
        }
        return processed_instance

    def _generate_processed_instance(self, raw_instance: dict) -> RagasDataPoint:
        """
        Generates a processed instance for the raw dataset.
        """
        # Check for the type of columns in raw_instance
        if self.col_query in raw_instance and not isinstance(raw_instance.get(self.col_query), str):
            raise TypeError(f"'{self.col_query}' is not of type string.")
        if self.col_contexts in raw_instance:
            if not isinstance(raw_instance.get(self.col_contexts), list):
                raise TypeError(f"'{self.col_contexts}' is not of type list.")
            if not all(isinstance(element, str) for element in raw_instance.get(self.col_contexts)):
                raise TypeError(f"Not all elements in '{self.col_context}' are of type string.")
        if self.col_response in raw_instance and not isinstance(raw_instance.get(self.col_response), str):
            raise TypeError(f"'{self.col_response}' is not of type string.")
        if self.col_expected_response in raw_instance and not isinstance(raw_instance.get(self.col_expected_response), str):
            raise TypeError(f"'{self.col_expected_response}' is not of type string.")

        # Create a processed instance
        processed_instance = {
            "query": raw_instance.get(self.col_query, None),
            "contexts": raw_instance.get(self.col_contexts, None),
            "response": raw_instance.get(self.col_response, None),
            "expected_response": raw_instance.get(self.col_expected_response, None)
        }
        return processed_instance

    def process(self) -> None:
        """
        Transforms the raw data into a structured format. Processes each entry from the raw dataset, and extracts attributes.

        Raises:
            TypeError: If any key is not of expected type.
        """
        if self.query_engine is not None:
            for raw_instance in self._raw_dataset:
                processed_instance = self._generate_processed_instance_for_llama_index(raw_instance)
                self._processed_dataset.append(processed_instance)
        else:
            for raw_instance in self._raw_dataset:
                processed_instance = self._generate_processed_instance(raw_instance)
                self._processed_dataset.append(processed_instance)

    def load_athina_inferences(
        self,
        filters: Optional[AthinaFilters] = None,
        limit: int = 10,
        context_key: Optional[str] = None,
    ):
        """
        Load data from Athina API.
        By default, this will fetch the last 10 inferences from the API.
        """
        pass